# noqa: F401

import logging
import os

import numpy as np

try:
    if os.environ.get("CUTDE_USE_BACKEND", "cuda") != "cuda":
        # The CUTDE_USE_BACKEND environment variable overrides the normal
        # choice of backend.
        # This can be helpful for testing purposes when it might be nice to run
        # with OpenCL or C++ even if CUDA is installed.
        raise ImportError

    from cutde.cuda import (  # noqa: F401
        empty,
        get,
        load_module,
        max_block_size,
        to,
        zeros,
    )

    which_backend = "cuda"
except ImportError:
    try:
        if os.environ.get("CUTDE_USE_BACKEND", "opencl") != "opencl":
            raise ImportError
        from cutde.opencl import (  # noqa: F401
            empty,
            get,
            load_module,
            max_block_size,
            to,
            zeros,
        )

        which_backend = "opencl"

    except ImportError:
        from cutde.cpp import (  # noqa: F401
            empty,
            get,
            load_module,
            max_block_size,
            to,
            zeros,
        )

        which_backend = "cpp"

logger = logging.getLogger(__name__)
logger.debug(f'cutde is using the "{which_backend}" backend')


def np_to_c_type(t):
    if t == np.float32:
        return "float"
    elif t == np.float64:
        return "double"


def intervals(length, step_size):
    out = []
    next_start = 0
    next_end = step_size
    while next_end < length + step_size:
        this_end = min(next_end, length)
        out.append((next_start, this_end))
        next_start += step_size
        next_end += step_size
    return out



# Modified from code presented in:
# Nikkhoo, M., Walter, T. R. (2015): Triangular dislocation: an analytical,
# artefact-free solution. - Geophysical Journal International, 201,
# 1117-1139. doi: 10.1093/gji/ggv035

# Original documentation:
# TDdispFS
# calculates displacements associated with a triangular dislocation in an
# elastic full-space.
#
# TD: Triangular Dislocation
# EFCS: Earth-Fixed Coordinate System
# TDCS: Triangular Dislocation Coordinate System
# ADCS: Angular Dislocation Coordinate System
#
# INPUTS
# X, Y and Z:
# Coordinates of calculation points in EFCS (East, North, Up). X, Y and Z
# must have the same size.
#
# P1,P2 and P3:
# Coordinates of TD vertices in EFCS.
#
# Ss, Ds and Ts:
# TD slip vector components (Strike-slip, Dip-slip, Tensile-slip).
#
# nu:
# Poisson's ratio.
#
# OUTPUTS
# ue, un and uv:
# Calculated displacement vector components in EFCS. ue, un and uv have
# the same unit as Ss, Ds and Ts in the inputs.
#
# Original documentation license:
# Copyright (c) 2014 Mehdi Nikkhoo
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the
# following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np


def build_tri_coordinate_system(tri):
    Vnorm = normalize(np.cross(tri[1] - tri[0], tri[2] - tri[0]))
    eY = np.array([0, 1, 0])
    eZ = np.array([0, 0, 1])
    Vstrike = np.cross(eZ, Vnorm)
    if np.linalg.norm(Vstrike) == 0:
        Vstrike = eY * Vnorm[2]
    Vstrike = normalize(Vstrike)
    Vdip = np.cross(Vnorm, Vstrike)
    return np.array([Vnorm, Vstrike, Vdip])


def normalize(v):
    return v / np.linalg.norm(v)


def trimodefinder(obs, tri):
    # trimodefinder calculates the normalized barycentric coordinates of
    # the points with respect to the TD vertices and specifies the appropriate
    # artefact-free configuration of the angular dislocations for the
    # calculations. The input matrices x, y and z share the same size and
    # correspond to the y, z and x coordinates in the TDCS, respectively. p1,
    # p2 and p3 are two-component matrices representing the y and z coordinates
    # of the TD vertices in the TDCS, respectively.
    # The components of the output (trimode) corresponding to each calculation
    # points, are 1 for the first configuration, -1 for the second
    # configuration and 0 for the calculation point that lie on the TD sides.

    a = (
        (tri[1, 1] - tri[2, 1]) * (obs[0] - tri[2, 0])
        + (tri[2, 0] - tri[1, 0]) * (obs[1] - tri[2, 1])
    ) / (
        (tri[1, 1] - tri[2, 1]) * (tri[0, 0] - tri[2, 0])
        + (tri[2, 0] - tri[1, 0]) * (tri[0, 1] - tri[2, 1])
    )
    b = (
        (tri[2, 1] - tri[0, 1]) * (obs[0] - tri[2, 0])
        + (tri[0, 0] - tri[2, 0]) * (obs[1] - tri[2, 1])
    ) / (
        (tri[1, 1] - tri[2, 1]) * (tri[0, 0] - tri[2, 0])
        + (tri[2, 0] - tri[1, 0]) * (tri[0, 1] - tri[2, 1])
    )
    c = 1 - a - b

    result = 1
    if (
        (a <= 0 and b > c and c > a)
        or (b <= 0 and c > a and a > b)
        or (c <= 0 and a > b and b > c)
    ):
        result = -1
    if (
        (a == 0 and b >= 0 and c >= 0)
        or (a >= 0 and b == 0 and c >= 0)
        or (a >= 0 and b >= 0 and c == 0)
    ):
        result = 0
    if result == 0 and obs[2] != 0:
        result = 1
    return result


def TDSetupD(obs, alpha, slip_b, nu, TriVertex, SideVec):
    # TDSetupD transforms coordinates of the calculation points as well as
    # slip vector components from ADCS into TDCS. It then calculates the
    # displacements in ADCS and transforms them into TDCS.

    # Transformation matrix
    A = np.array([[SideVec[2], -SideVec[1]], [SideVec[1], SideVec[2]]])

    # Transform coordinates of the calculation points from TDCS into ADCS
    r1 = A.dot([obs[1] - TriVertex[1], obs[2] - TriVertex[2]])
    y1 = r1[0]
    z1 = r1[1]

    # Transform the in-plane slip vector components from TDCS into ADCS
    r2 = A.dot([slip_b[1], slip_b[2]])
    by1 = r2[0]
    bz1 = r2[1]

    # Calculate displacements associated with an angular dislocation in ADCS
    [u, v0, w0] = AngDisDisp(obs[0], y1, z1, -np.pi + alpha, slip_b[0], by1, bz1, nu)

    # Transform displacements from ADCS into TDCS
    r3 = A.T.dot([v0, w0])
    v = r3[0]
    w = r3[1]
    return u, v, w


def AngDisDisp(x, y, z, alpha, bx, by, bz, nu):
    # AngDisDisp calculates the "incomplete" displacements (without the
    # Burgers' function contribution) associated with an angular dislocation in
    # an elastic full-space.

    cosA = np.cos(alpha)
    sinA = np.sin(alpha)
    eta = y * cosA - z * sinA
    zeta = y * sinA + z * cosA
    r = np.sqrt(x**2 + y**2 + z**2)

    # Avoid complex results for the logarithmic terms
    if zeta > r:
        zeta = r
    if z > r:
        z = r

    ux = bx / 8 / np.pi / (1 - nu) * (x * y / r / (r - z) - x * eta / r / (r - zeta))
    vx = (
        bx
        / 8
        / np.pi
        / (1 - nu)
        * (
            eta * sinA / (r - zeta)
            - y * eta / r / (r - zeta)
            + y**2 / r / (r - z)
            + (1 - 2 * nu) * (cosA * np.log(r - zeta) - np.log(r - z))
        )
    )
    wx = (
        bx
        / 8
        / np.pi
        / (1 - nu)
        * (
            eta * cosA / (r - zeta)
            - y / r
            - eta * z / r / (r - zeta)
            - (1 - 2 * nu) * sinA * np.log(r - zeta)
        )
    )

    uy = (
        by
        / 8
        / np.pi
        / (1 - nu)
        * (
            x**2 * cosA / r / (r - zeta)
            - x**2 / r / (r - z)
            - (1 - 2 * nu) * (cosA * np.log(r - zeta) - np.log(r - z))
        )
    )
    vy = (
        by
        * x
        / 8
        / np.pi
        / (1 - nu)
        * (y * cosA / r / (r - zeta) - sinA * cosA / (r - zeta) - y / r / (r - z))
    )
    wy = (
        by
        * x
        / 8
        / np.pi
        / (1 - nu)
        * (z * cosA / r / (r - zeta) - cosA**2 / (r - zeta) + 1 / r)
    )

    uz = (
        bz
        * sinA
        / 8
        / np.pi
        / (1 - nu)
        * ((1 - 2 * nu) * np.log(r - zeta) - x**2 / r / (r - zeta))
    )
    vz = bz * x * sinA / 8 / np.pi / (1 - nu) * (sinA / (r - zeta) - y / r / (r - zeta))
    wz = bz * x * sinA / 8 / np.pi / (1 - nu) * (cosA / (r - zeta) - z / r / (r - zeta))

    return ux + uy + uz, vx + vy + vz, wx + wy + wz


def TDdispFS(obs, tri, slip, nu):
    transform = build_tri_coordinate_system(tri)

    transformed_obs = transform.dot(obs - tri[1])
    transformed_tri = np.zeros((3, 3))
    transformed_tri[0, :] = transform.dot(tri[0] - tri[1])
    transformed_tri[2, :] = transform.dot(tri[2] - tri[1])
    np.testing.assert_almost_equal(transformed_tri[1], [0, 0, 0])
    np.testing.assert_almost_equal(transformed_tri[0][0], 0)
    np.testing.assert_almost_equal(transformed_tri[2][0], 0)

    e12 = normalize(transformed_tri[1] - transformed_tri[0])
    e13 = normalize(transformed_tri[2] - transformed_tri[0])
    e23 = normalize(transformed_tri[2] - transformed_tri[1])

    A = np.arccos(e12.T.dot(e13))
    B = np.arccos(-e12.T.dot(e23))
    C = np.arccos(e23.T.dot(e13))

    mode = trimodefinder(
        np.array([transformed_obs[1], transformed_obs[2], transformed_obs[0]]),
        transformed_tri[:, 1:],
    )

    slip_b = np.array([slip[2], slip[0], slip[1]])
    if mode == 1:
        # Calculate first angular dislocation contribution
        u1Tp, v1Tp, w1Tp = TDSetupD(
            transformed_obs, A, slip_b, nu, transformed_tri[0], -e13
        )
        # Calculate second angular dislocation contribution
        u2Tp, v2Tp, w2Tp = TDSetupD(
            transformed_obs, B, slip_b, nu, transformed_tri[1], e12
        )
        # Calculate third angular dislocation contribution
        u3Tp, v3Tp, w3Tp = TDSetupD(
            transformed_obs, C, slip_b, nu, transformed_tri[2], e23
        )
        out = np.array([u1Tp + u2Tp + u3Tp, v1Tp + v2Tp + v3Tp, w1Tp + w2Tp + w3Tp])
    elif mode == -1:
        # Calculate first angular dislocation contribution
        u1Tn, v1Tn, w1Tn = TDSetupD(
            transformed_obs, A, slip_b, nu, transformed_tri[0], e13
        )
        # Calculate second angular dislocation contribution
        u2Tn, v2Tn, w2Tn = TDSetupD(
            transformed_obs, B, slip_b, nu, transformed_tri[1], -e12
        )
        # Calculate third angular dislocation contribution
        u3Tn, v3Tn, w3Tn = TDSetupD(
            transformed_obs, C, slip_b, nu, transformed_tri[2], -e23
        )
        out = np.array([u1Tn + u2Tn + u3Tn, v1Tn + v2Tn + v3Tn, w1Tn + w2Tn + w3Tn])
    else:
        out = np.array([np.nan, np.nan, np.nan])

    a = np.array(
        [
            -transformed_obs[0],
            transformed_tri[0][1] - transformed_obs[1],
            transformed_tri[0][2] - transformed_obs[2],
        ]
    )
    b = -transformed_obs
    c = np.array(
        [
            -transformed_obs[0],
            transformed_tri[2][1] - transformed_obs[1],
            transformed_tri[2][2] - transformed_obs[2],
        ]
    )
    na = np.sqrt(np.sum(a**2))
    nb = np.sqrt(np.sum(b**2))
    nc = np.sqrt(np.sum(c**2))

    FiN = (
        a[0] * (b[1] * c[2] - b[2] * c[1])
        - a[1] * (b[0] * c[2] - b[2] * c[0])
        + a[2] * (b[0] * c[1] - b[1] * c[0])
    )
    FiD = na * nb * nc + np.sum(a * b) * nc + np.sum(a * c) * nb + np.sum(b * c) * na
    Fi = -2 * np.arctan2(FiN, FiD) / 4 / np.pi

    # Calculate the complete displacement vector components in TDCS
    out += slip_b * Fi

    # Transform the complete displacement vector components from TDCS into EFCS
    return transform.T.dot(out)



import os
import warnings
from math import ceil

import numpy as np

import cutde.backend as backend

source_dir = os.path.dirname(os.path.realpath(__file__))

DISP_FS = ("disp_fs", 3)
STRAIN_FS = ("strain_fs", 6)
DISP_HS = ("disp_hs", 3)
STRAIN_HS = ("strain_hs", 6)


class Placeholder:
    pass


placeholder = Placeholder()


def solve_types(obs_pts, tris, slips):
    type_map = {
        np.int32: np.float32,
        np.int64: np.float64,
        np.float32: np.float32,
        np.float64: np.float64,
    }

    float_type = None
    out_arrs = []
    for name, arr in [("obs_pts", obs_pts), ("tris", tris), ("slips", slips)]:
        if isinstance(arr, Placeholder):
            out_arrs.append(arr)
            continue

        dtype = arr.dtype.type

        if dtype not in type_map:
            raise ValueError(
                f"The {name} input array has type {arr.dtype} but must have a float or"
                " integer dtype."
            )

        if float_type is None:
            float_type = type_map[dtype]

            # If we're using OpenCL, we need to check if float64 is allowed.
            # If not, convert to float32.
            if backend.which_backend == "opencl":
                import cutde.opencl

                cutde.opencl.ensure_initialized()
                extensions = (
                    cutde.opencl.gpu_ctx.devices[0].extensions.strip().split(" ")
                )
                if "cl_khr_fp64" not in extensions and float_type is np.float64:
                    warnings.warn(
                        "The OpenCL implementation being used does not support "
                        "float64. This will require converting arrays to float32."
                    )
                    float_type = np.float32

        out_arr = arr
        if dtype != float_type:
            warnings.warn(
                f"The {name} input array has type {out_arr.dtype} but needs "
                "to be converted"
                f" to dtype {np.dtype(float_type)}. Converting {name} to "
                f"{np.dtype(float_type)} may be expensive."
            )
            out_arr = out_arr.astype(float_type)

        if out_arr.flags.f_contiguous:
            warnings.warn(
                f"The {name} input array has Fortran ordering. "
                "Converting to C ordering. This may be expensive."
            )
            out_arr = np.ascontiguousarray(out_arr)

        out_arrs.append(out_arr)

    return float_type, out_arrs


def check_inputs(obs_pts, tris, slips):
    if obs_pts.shape[1] != 3:
        raise ValueError(
            "The second dimension of the obs_pts array must be 3 because the "
            "observation points should be locations in three-dimensional space."
        )
    if tris.shape[1] != 3:
        raise ValueError(
            "The second dimension of the tris array must be 3 because there must be "
            "three vertices per triangle."
        )
    if tris.shape[2] != 3:
        raise ValueError(
            "The third dimension of the tris array must be 3 because the triangle "
            "vertices should be locations in three-dimensional space."
        )
    if not isinstance(slips, Placeholder) and (slips.shape[0] != tris.shape[0]):
        raise ValueError(
            "The number of input slip vectors must be equal to the number of input"
            " triangles."
        )
    if not isinstance(slips, Placeholder) and (slips.shape[1] != 3):
        raise ValueError(
            "The second dimension of the slips array must be 3 because each row "
            "should be a vector in the TDE coordinate system (strike-slip, dip-slip,"
            " tensile-slip)."
        )


def call_clu(obs_pts, tris, slips, nu, fnc):
    fnc_name, vec_dim = fnc
    if tris.shape[0] != obs_pts.shape[0]:
        raise ValueError("There must be one input observation point per triangle.")

    check_inputs(obs_pts, tris, slips)
    float_type, (obs_pts, tris, slips) = solve_types(obs_pts, tris, slips)

    n = obs_pts.shape[0]
    block_size = backend.max_block_size(16)
    n_blocks = int(np.ceil(n / block_size))
    gpu_config = dict(
        block_size=block_size, float_type=backend.np_to_c_type(float_type)
    )
    module = backend.load_module("pairs.cu", tmpl_args=gpu_config, tmpl_dir=source_dir)

    gpu_results = backend.empty(n * vec_dim, float_type)
    gpu_obs_pts = backend.to(obs_pts, float_type)
    gpu_tris = backend.to(tris, float_type)
    gpu_slips = backend.to(slips, float_type)

    getattr(module, "pairs_" + fnc_name)(
        gpu_results,
        np.int32(n),
        gpu_obs_pts,
        gpu_tris,
        gpu_slips,
        float_type(nu),
        (n_blocks, 1, 1),
        (block_size, 1, 1),
    )
    out = backend.get(gpu_results).reshape((n, vec_dim))
    return out


def call_clu_matrix(obs_pts, tris, nu, fnc):
    fnc_name, vec_dim = fnc
    check_inputs(obs_pts, tris, placeholder)
    float_type, (obs_pts, tris, _) = solve_types(obs_pts, tris, placeholder)

    n_obs = obs_pts.shape[0]
    n_src = tris.shape[0]
    block_size = backend.max_block_size(16)
    n_obs_blocks = int(np.ceil(n_obs / block_size))
    n_src_blocks = int(np.ceil(n_src / block_size))
    gpu_config = dict(float_type=backend.np_to_c_type(float_type))
    module = backend.load_module("matrix.cu", tmpl_args=gpu_config, tmpl_dir=source_dir)

    gpu_results = backend.empty(n_obs * vec_dim * n_src * 3, float_type)
    gpu_obs_pts = backend.to(obs_pts, float_type)
    gpu_tris = backend.to(tris, float_type)

    getattr(module, "matrix_" + fnc_name)(
        gpu_results,
        np.int32(n_obs),
        np.int32(n_src),
        gpu_obs_pts,
        gpu_tris,
        float_type(nu),
        (n_obs_blocks, n_src_blocks, 1),
        (block_size, block_size, 1),
    )
    out = backend.get(gpu_results).reshape((n_obs, vec_dim, n_src, 3))
    return out


def call_clu_free(obs_pts, tris, slips, nu, fnc):
    fnc_name, vec_dim = fnc
    check_inputs(obs_pts, tris, slips)
    float_type, (obs_pts, tris, slips) = solve_types(obs_pts, tris, slips)

    n_obs = obs_pts.shape[0]
    n_src = tris.shape[0]
    block_size = backend.max_block_size(256)

    gpu_obs_pts = backend.to(obs_pts, float_type)
    gpu_tris = backend.to(tris, float_type)
    gpu_slips = backend.to(slips, float_type)
    gpu_results = backend.zeros(n_obs * vec_dim, float_type)

    n_obs_blocks = int(np.ceil(n_obs / block_size))
    gpu_config = dict(float_type=backend.np_to_c_type(float_type))
    module = backend.load_module("free.cu", tmpl_args=gpu_config, tmpl_dir=source_dir)

    # Split up the sources into chunks so that we don't completely overwhelm a
    # single GPU machine and cause the screen to lock up.
    default_chunk_size = 64
    n_chunks = int(ceil(n_src / default_chunk_size))
    out = np.zeros((n_obs, vec_dim), dtype=float_type)
    for i in range(n_chunks):
        chunk_start = i * default_chunk_size
        chunk_size = min(n_src - chunk_start, default_chunk_size)
        chunk_end = chunk_start + chunk_size

        getattr(module, "free_" + fnc_name)(
            gpu_results,
            np.int32(n_obs),
            np.int32(n_src),
            np.int32(chunk_start),
            np.int32(chunk_end),
            gpu_obs_pts,
            gpu_tris,
            gpu_slips,
            float_type(nu),
            (n_obs_blocks, 1, 1),
            (block_size, 1, 1),
        )
        out += backend.get(gpu_results).reshape((n_obs, vec_dim))
    return out


def process_block_inputs(obs_start, obs_end, src_start, src_end):
    out_arrs = []
    for name, a in [
        ("obs_start", obs_start),
        ("obs_end", obs_end),
        ("src_start", src_start),
        ("src_end", src_end),
    ]:
        a = np.array(a)
        if a.shape[0] != np.array(obs_start).shape[0]:
            raise ValueError(f"The length of {name} must match obs_start.")
        if not (a.dtype.type is np.int32 or a.dtype.type is np.int64):
            raise ValueError(f"The {name} array must have integer type.")

        # Don't bother warning for these conversions since the cost of
        # converting a single value per block is tiny.
        if a.flags.f_contiguous or a.dtype.type != np.int32:
            a = np.ascontiguousarray(a, dtype=np.int32)

        out_arrs.append(a)

    return out_arrs


def call_clu_block(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, fnc):
    fnc_name, vec_dim = fnc
    check_inputs(obs_pts, tris, placeholder)
    float_type, (obs_pts, tris, _) = solve_types(obs_pts, tris, placeholder)
    obs_start, obs_end, src_start, src_end = process_block_inputs(
        obs_start, obs_end, src_start, src_end
    )

    block_sizes = vec_dim * 3 * (obs_end - obs_start) * (src_end - src_start)
    block_end = np.cumsum(block_sizes)
    block_start = np.empty(block_end.shape[0] + 1, dtype=block_end.dtype)
    block_start[:-1] = block_end - block_sizes
    block_start[-1] = block_end[-1]

    n_blocks = obs_end.shape[0]
    team_size = backend.max_block_size(16)
    gpu_config = dict(float_type=backend.np_to_c_type(float_type))
    module = backend.load_module("blocks.cu", tmpl_args=gpu_config, tmpl_dir=source_dir)

    gpu_results = backend.zeros(block_end[-1], float_type)
    gpu_obs_pts = backend.to(obs_pts, float_type)
    gpu_tris = backend.to(tris, float_type)
    gpu_obs_start = backend.to(obs_start, np.int32)
    gpu_obs_end = backend.to(obs_end, np.int32)
    gpu_src_start = backend.to(src_start, np.int32)
    gpu_src_end = backend.to(src_end, np.int32)
    gpu_block_start = backend.to(block_start, np.int32)

    getattr(module, "blocks_" + fnc_name)(
        gpu_results,
        gpu_obs_pts,
        gpu_tris,
        gpu_obs_start,
        gpu_obs_end,
        gpu_src_start,
        gpu_src_end,
        gpu_block_start,
        float_type(nu),
        (n_blocks, 1, 1),
        (team_size, 1, 1),
    )
    return backend.get(gpu_results), block_start
